/* 
   conn.c - connections handling

   Copyright (C) Szymon Acedanski 2003
*/

/*
    ggrelay - Gadu-Gadu instant messager relaying agent
    Copyright (C) 2003 Szymon Acedanski

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>

#include "ggrelay.h"

static struct gg_server_desc* select_server(void)
{
	struct gg_server_desc* svr = gg_servers_iter;
	if (!(gg_servers_iter = gg_servers_iter->next))
		gg_servers_iter = gg_servers;
	return svr;
}

static void block_signals(void)
{
	sigset_t s;
	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK, &s, NULL);
}

static void gg_cleanup_fd(struct gg_fd* f)
{
	if (f->state == GG_S_DESTROY) return;
#ifdef HAVE_OPENSSL
	if (f->flags & GG_FLAG_SSL)
	{
		gg_ssl_shutdown(f, SHUT_RDWR);
		gg_ssl_close(f);
	}
	else
#endif
	{
		shutdown(f->fd, SHUT_RDWR);
		close(f->fd);
	}
	gg_masq_close_handler(f);
	f->state = GG_S_DESTROY;
	gg_assert(!list_empty(&f->list));
	if (f->type == GG_T_CHATCOUNTED)
		gg_max_chat_conns++;
	else if (f->type == GG_T_DCCCOUNTED)
		gg_max_dcc_conns++;
	if (f->parent)
	{
		gg_put_fd(f->parent);
		f->parent = NULL;
	}
	if (f->cexp)
	{
		list_del(&f->cexp->list);
		free(f->cexp);
		f->cexp = NULL;
	}
	gg_condlog("closed fd %d", f->fd);
}

void gg_close_fd(struct gg_fd* f)
{
	DEBUGP("closing fd %d", f->fd);
	if (f->expby)
	{
		gg_cleanup_fd(f->expby);
		f->expby = NULL;
	}
	if (f->twin)
	{
		if (f->readoffset)
		{
#ifdef HAVE_OPENSSL
			if (f->twin->flags & GG_FLAG_SSL)
				gg_ssl_shutdown(f->twin, SHUT_RD);
			else
#endif
				shutdown(f->twin->fd, SHUT_RD);
			f->twin->state = GG_S_CLOSING;
			f->twin->twin = f->twin;
			memcpy(f->twin->buffer, f->buffer, f->readoffset);
			f->twin->readoffset = f->readoffset;
			f->twin->writeoffset = f->readoffset;
		}
		else
			gg_cleanup_fd(f->twin);
	}
	gg_cleanup_fd(f);
}

void gg_abort_fd(struct gg_fd* f)
{
	DEBUGP("aborting fd %d", f->fd);
	gg_cleanup_fd(f);
	if (f->twin)
		gg_cleanup_fd(f->twin);
}

static inline void gg_adjust_sockbuffs(int fd)
{
	int foo = GG_SOCK_BUFSIZE;
	setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &foo, sizeof(int));
	setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &foo, sizeof(int));
}

static void* gg_accept_chat(void* __info)
{
	struct gg_connect_info* _info = (struct gg_connect_info*)__info;
	struct gg_connect_info info = *_info;
	struct gg_server_desc* svr;
	struct sockaddr_in sin;
	int fd = -1;
	int foo = 1;
	struct gg_fd* f;
	free(__info);
	
	block_signals();
	
#ifdef HAVE_OPENSSL
	if (info.twin->flags & GG_FLAG_SSL)
		if (gg_ssl_accept_connection(info.twin))
			goto failure;
#endif

	if (ioctl(info.twin->fd, FIONBIO, &foo, sizeof(foo)))
	{
		DEBUGP("ioctl(FIONBIO) failed");
		goto failure;
	}
	
	pthread_mutex_lock(&gg_servers_mutex);
	svr = select_server();
	pthread_mutex_unlock(&gg_servers_mutex);

	if ((fd = socket(PF_INET, SOCK_STREAM, 0)) == -1)
	{
		DEBUGP("socket() failed");
		goto failure;
	}

	gg_adjust_sockbuffs(fd);
	sin.sin_family = AF_INET;
	sin.sin_port = 0;
	sin.sin_addr.s_addr = INADDR_ANY;
	if (bind(fd, (struct sockaddr*)&sin, sizeof(sin)))
	{
		DEBUGP("bind() failed");
		goto closefd;
	}
	DEBUGP("conn to %u.%u.%u.%u:%hd", NIPQUAD(svr->addr.sin_addr.s_addr),
			ntohs(svr->addr.sin_port));
	if (connect(fd, (struct sockaddr*)&svr->addr, sizeof(svr->addr)))
	{
		DEBUGP("connect() failed (errno=%d)", errno);
		goto closefd;
	}

	if (!(f = gg_alloc_fd(fd, gg_dynamic_buffers ?
					GG_CHATREPEATER_BUFSIZE_DYNAMIC :
					GG_CHATREPEATER_BUFSIZE_STATIC)))
		goto closefd;
	f->type = GG_T_NOTCOUNTED;
	f->twin = info.twin;
	f->fd = fd;
	info.twin->twin = f;
	gg_masq_init_chat_req(info.twin);
	gg_masq_init_chat_resp(f);
	
	gg_condlog("established chat connection on fds %d:%d",
			f->twin->fd, f->fd);
	
	pthread_mutex_lock(&gg_fds_mutex);
	__gg_attach_fd(f);
	__gg_attach_fd(info.twin);
	pthread_mutex_unlock(&gg_fds_mutex);
	gg_kick_main_thread();

	return (void*)0;
	
closefd:
	close(fd);
failure:
	gg_condlog("failure establishing chat connection on %d:%d",
			info.twin->fd, fd);
	pthread_mutex_lock(&gg_fds_mutex);
	gg_put_fd(info.twin); /* not attached, must be put explicitly */
	pthread_mutex_unlock(&gg_fds_mutex);
	return (void*)1;
}

void* gg_establish_dcc(void* __info)
{
	struct gg_connect_info* _info = (struct gg_connect_info*)__info;
	struct gg_connect_info info = *_info;
	struct sockaddr_in sin;
	int fd;
	struct gg_fd* f;
	free(__info);
	
	block_signals();
	
	if ((fd = socket(PF_INET, SOCK_STREAM, 0)) == -1)
		goto failure;

	gg_adjust_sockbuffs(fd);
	sin.sin_family = AF_INET;
	sin.sin_port = 0;
	sin.sin_addr.s_addr = INADDR_ANY;
	if (bind(fd, (struct sockaddr*)&sin, sizeof(sin)))
		goto closefd;

	sin.sin_port = htons(info.nat_port);
	sin.sin_addr.s_addr = info.nat_ip;
	if (connect(fd, (struct sockaddr*)&sin, sizeof(sin)))
		goto closefd;

	if (!(f = gg_alloc_fd(fd, GG_DCCREPEATER_BUFSIZE)))
		goto closefd;
	f->type = GG_T_NOTCOUNTED;
	f->state = GG_S_REPEATMODE;
	f->twin = info.twin;
	f->fd = fd;

	gg_condlog("established dcc connection on fds %d:%d",
			f->twin->fd, f->fd);

	pthread_mutex_lock(&gg_fds_mutex);
	info.twin->twin = f;
	info.twin->state = GG_S_REPEATMODE;
	if (info.twin->state == GG_S_DESTROY)
	{
		pthread_mutex_unlock(&gg_fds_mutex);
		DEBUGP("caught just-destroyed condition on %d:%d",
				info.twin->fd, fd);
		gg_put_fd(f);
		gg_put_fd(info.twin);
		close(fd);
		return (void*)1;
	}
	__gg_attach_fd(f);
	pthread_mutex_unlock(&gg_fds_mutex);
	gg_kick_main_thread();
	gg_put_fd(info.twin);

	return (void*)0;
	
closefd:
	close(fd);
failure:
	gg_condlog("failure establishing dcc connection on %d:%d",
			info.twin->fd, fd);
	pthread_mutex_lock(&gg_fds_mutex);
	gg_cleanup_fd(info.twin);
	pthread_mutex_unlock(&gg_fds_mutex);
	gg_put_fd(info.twin);
	return (void*)1;
}

int gg_accept_chat_connection(int lfd, int ssl)
{
	struct sockaddr_in sin;
	socklen_t addrlen = sizeof(sin);
	int fd = accept(lfd, (struct sockaddr*)&sin, &addrlen);
	struct gg_fd* f;
	struct gg_connect_info* info;
	pthread_attr_t attr;
	pthread_t dummy;

	/* gg_fds_mutex is locked. beware. */

	if (fd == -1)
		return -1;

	if (!gg_max_chat_conns)
	{
		gg_condlog("REJECTED chat connection fd %d from %u.%u.%u.%u:%u",
				fd, NIPQUAD(sin.sin_addr), ntohs(sin.sin_port));
		goto closefd;
	}

	gg_condlog("accepted chat connection fd %d from %u.%u.%u.%u:%u",
			fd, NIPQUAD(sin.sin_addr), ntohs(sin.sin_port));
	
	if (!(f = gg_alloc_fd(fd, gg_dynamic_buffers ?
					GG_CHATREPEATER_BUFSIZE_DYNAMIC :
					GG_CHATREPEATER_BUFSIZE_STATIC)))
		goto closefd;
	gg_adjust_sockbuffs(fd);
	f->ip = sin.sin_addr.s_addr;
	if (!(info = (struct gg_connect_info*)malloc(sizeof(*info))))
		goto freef;
	f->type = GG_T_CHATCOUNTED;
	f->flags |= GG_FLAG_CLIENTCONN;
	if (ssl)
		f->flags |= GG_FLAG_SSL;
	info->twin = f;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	DEBUGP("spawning helper thread...");
	if (pthread_create(&dummy, &attr, gg_accept_chat, info))
		goto freeinfo;
	gg_max_chat_conns--;
	return 0;
	
freeinfo:
	free(info);
freef:
	gg_put_fd(f);
closefd:
	close(fd);
	return -1;
}

int gg_accept_dcc_connection(void)
{
	struct sockaddr_in sin;
	socklen_t addrlen = sizeof(sin);
	int fd = accept(gg_dcc_listener, (struct sockaddr*)&sin, &addrlen);
	struct gg_fd* f;

	/* gg_fds_mutex is locked. beware. */
	
	if (fd == -1)
		return -1;

	if (!gg_max_dcc_conns)
	{
		gg_condlog("REJECTED dcc connection fd %d from %u.%u.%u.%u:%u",
				fd, NIPQUAD(sin.sin_addr), ntohs(sin.sin_port));
		goto closefd;
	}

	gg_condlog("accepted dcc connection fd %d from %u.%u.%u.%u:%u",
			fd, NIPQUAD(sin.sin_addr), ntohs(sin.sin_port));
	
	if (!(f = gg_alloc_fd(fd, GG_DCCREPEATER_BUFSIZE)))
		goto closefd;
	gg_adjust_sockbuffs(fd);
	f->ip = sin.sin_addr.s_addr;
	f->type = GG_T_DCCCOUNTED;
	gg_masq_init_dcc_req(f);
	__gg_attach_fd(f);
	gg_max_dcc_conns--;
	return 0;
	
closefd:
	close(fd);
	return -1;
}

int gg_need_dcc_server(void)
{
	uint32_t newip;
	
	DEBUGP("gg_outer_iface=%s gg_outer_ip=%u.%u.%u.%u",
			gg_outer_iface, NIPQUAD(gg_outer_ip));
	
	if (!gg_outer_iface) return 0;
	newip = gg_ifname_to_ip(gg_outer_iface);
	if (newip == gg_outer_ip)
		return !newip;
	gg_outer_ip = newip;

	/* must reestablish */

	gg_condlog("(re)binding dcc listener to %u.%u.%u.%u", NIPQUAD(newip));
	if (gg_dcc_listener >= 0)
		close(gg_dcc_listener);
	if ((gg_dcc_listener = gg_setup_listener(gg_outer_ip, gg_dcc_port)) < 0)
	{
		gg_outer_ip = 0;
		return 1;
	}
	return 0;
}

int gg_setup_listener(uint32_t ip, uint16_t port)
{
	int fd = socket(PF_INET, SOCK_STREAM, 0);
	struct sockaddr_in sin;
	int dummy = 1;

	gg_condlog("setting up listener on %u.%u.%u.%u:%u", NIPQUAD(ip), port);
	
	if (fd == -1)
	{
		gg_error("socket() failed: %s", strerror(errno));
		return -1;
	}
	
	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = ip;

	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char*)&dummy, sizeof(dummy)))
	{
		gg_error("setsockopt() failed: %s", strerror(errno));
		goto closesock;
	}

	if (bind(fd, (struct sockaddr*)&sin, sizeof(sin)))
	{
		gg_error("bind() failed: %s", strerror(errno));
		goto closesock;
	}

	if (listen(fd, GG_LISTENER_BACKLOG))
	{
		gg_error("listen() failed: %s", strerror(errno));
		goto closesock;
	}
	return fd;
	
closesock:
	close(fd);
	return -1;
}

int gg_fd_wants_reading(struct gg_fd* f)
{
	if (f->flags & GG_FLAG_READFORWRITING) return 1;
	if (f->readoffset == f->bufsize) return 0;
	if (f->state == GG_S_CLOSING) return 0;
	if (f->state == GG_S_WRTCMPL) return 0;
	if (f->state == GG_S_SUSPENDED) return 0; /* is it necessary? */
	if (f->twin && f->twin->flags & GG_FLAG_WAITFORME) return 0;
	return 1;
}

int gg_fd_wants_writing(struct gg_fd* f)
{
	if (f->flags & GG_FLAG_WRITEFORREADING) return 1;
	if (f->state == GG_S_SUSPENDED) return 0; /* is it necessary? */
	if (!f->twin) return 0;
	if (f->twin->writeoffset == 0) return 0;
	if (f->twin->flags & GG_FLAG_WAITFORME) return 0;
	return 1;
}

static __inline__ void gg_check_expected(struct gg_fd* f)
{
	while (f->state == GG_S_EXPECTING && f->readoffset >= f->expectlen
			&& f->writeoffset == 0)
		f->expectfn(f);
}

void gg_fd_handle_reading(struct gg_fd* f)
{
	int size;

	if (f->state == GG_S_DESTROY) return;
	if (f->twin && f->twin != f && f->twin->flags & GG_FLAG_WAITFORME) return;

#ifdef HAVE_OPENSSL
	if (f->flags & GG_FLAG_SSL)
		size = gg_ssl_read(f);
	else
#endif
		size = read(f->fd, f->buffer + f->readoffset,
				f->bufsize - f->readoffset);

	DEBUGP("fd=%d %s twin=%d read=%d ro=%d wo=%d state=%d", f->fd,
			(f->flags & GG_FLAG_SSL) ? "(ssl)" : "", 
			f->twin?f->twin->fd:-1, size, f->readoffset, f->writeoffset, f->state);

	if (size == -2)
		return;
	else if (size == -1)
		gg_abort_fd(f);
	else if (size == 0)
		gg_close_fd(f);
	else
	{
		f->readoffset += size;
		if (f->state == GG_S_EXPECTING)
			gg_check_expected(f);
		else
			f->writeoffset = f->readoffset;
	}
}

void gg_fd_handle_writing(struct gg_fd* f)
{
	int written;

	if (f->state == GG_S_DESTROY) return;
	if (!f->twin)
	{
		gg_abort_fd(f);
		return;
	}

	if (f->state == GG_S_CLOSING && f->writeoffset == 0)
	{
		gg_close_fd(f);
		return;
	}

	if (f->twin != f && f->twin->flags & GG_FLAG_WAITFORME)
		return;
	
#ifdef HAVE_OPENSSL
	if (f->flags & GG_FLAG_SSL)
		written = gg_ssl_write(f);
	else
#endif
		written = write(f->fd, f->twin->buffer, f->twin->writeoffset);

	DEBUGP("fd=%d %s twin=%d written=%d ro=%d wo=%d state=%d", f->fd, 
			(f->flags & GG_FLAG_SSL) ? "(ssl)" : "", f->twin->fd, written,
			f->twin->readoffset, f->twin->writeoffset, f->state);

	if (written == -2)
		return;
	else if (written == -1)
		gg_abort_fd(f);
	else if (written == 0)
		gg_close_fd(f);
	else
	{
		f->twin->readoffset -= written;
		f->twin->writeoffset -= written;
		if (f->twin->readoffset != 0)
			memmove(f->twin->buffer, f->twin->buffer + written,
					f->twin->readoffset);
		if (f->twin->writeoffset == 0)
		{
			if (f->state == GG_S_CLOSING)
				gg_close_fd(f);
			else if (f->state == GG_S_WRTCMPL)
				f->expectfn(f);
			else
				gg_check_expected(f->twin);
		}
	}
}
