/* 
   ggrelay.h - some declarations and definitions

   Copyright (C) Szymon Acedanski 2003
*/

/*
    ggrelay - Gadu-Gadu instant messager relaying agent
    Copyright (C) 2003 Szymon Acedanski

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef GG_RELAY_H_INCLUDED
#define GG_RELAY_H_INCLUDED

#include <string.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <pthread.h>
#include <syslog.h>
#include <time.h>
#ifdef HAVE_OPENSSL
#include <openssl/ssl.h>
#endif

#ifndef __GNUC__
#define __inline__
#endif

#include "gglist.h"

#define GGRELAY_VERSION         "1.0-rc5A"

#define GG_DEFAULT_SERVER       "217.17.41.85:8074"
#define GG_DEFAULT_SERVER_PORT  8074
#define GG_DEFAULT_SSLSVR_PORT  8075
#define GG_DEFAULT_DCC_PORT     1550

#define GG_CHATREPEATER_BUFSIZE_STATIC    8192
#define GG_CHATREPEATER_BUFSIZE_DYNAMIC    512
#define GG_DCCREPEATER_BUFSIZE            2048

#define GG_SOCK_BUFSIZE         4096
#define GG_INJECT_BUFSIZE        512
#define GG_SHRINKBUF_THRESHOLD    20
#define GG_LISTENER_BACKLOG       10
#define GG_EXPECTATION_TIMEOUT   300 /* 5 minutes */
#define GG_MAXFD                8192

extern int gg_logging;
#define gg_condlog(args...) do { if (gg_logging) \
	syslog(LOG_NOTICE, ## args); } while (0)
#define gg_assert(cond)     do { if (!(cond)) syslog(LOG_ERR, \
	"assertion failed at %s:%d", __FILE__, __LINE__); } while (0)
#define gg_debug(args...)   syslog(LOG_DEBUG, ## args)
#define gg_notice(args...)  syslog(LOG_NOTICE, ## args)
#define gg_warn(args...)    syslog(LOG_WARNING, ## args)
#define gg_error(args...)   syslog(LOG_ERR, ## args)

#ifdef NDEBUG
#define DEBUGP(...)
#else
#define DEBUGP(fmt, args...) gg_debug("%s(): " fmt, __FUNCTION__, ## args)
#endif

#ifndef NIPQUAD
#define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
	((unsigned char *)&addr)[3]
#endif

struct gg_server_desc
{
	struct gg_server_desc* next;
	struct sockaddr_in addr;
};

struct gg_connect_info
{
	struct gg_fd* twin;
	uint32_t nat_ip;
	uint16_t nat_port;
};

struct gg_conn_expectation
{
	struct list_head list;
	uint32_t my_uin;
	uint32_t your_uin;
	struct gg_fd* twin;
	struct gg_fd* owner;
	time_t timeout;
};

enum gg_fd_type
{
	GG_T_INVALID,
	GG_T_CHATCOUNTED,
	GG_T_DCCCOUNTED,
	GG_T_NOTCOUNTED
};

enum gg_fd_state
{
	GG_S_EXPECTING,
	GG_S_REPEATMODE,
	GG_S_CLOSING,      /* twin fd no more valid, send data and say goodbye,
						  f->twin == f */
	GG_S_WRTCMPL,      /* set f->twin == f to get wait-for-write-completion */
	GG_S_SUSPENDED,
	GG_S_LISTENING,
	GG_S_DESTROY
};

struct gg_fd
{
	struct list_head list;
	int refcount;
	enum gg_fd_type type;
	enum gg_fd_state state;
	int fd;
	struct gg_fd* twin;
	struct gg_fd* parent;
	struct gg_conn_expectation* cexp;
	struct gg_fd* expby;
	uint32_t my_uin;
	uint32_t your_uin;
	uint16_t flags;
#ifdef HAVE_OPENSSL
	SSL* ssl;
#endif
	uint32_t ip;
	uint32_t uin;
	uint32_t nat_ip;
	uint16_t nat_port;
	int expectlen;
	void (*expectfn)(struct gg_fd*);
	int injectlen;
	char injectbuf[GG_INJECT_BUFSIZE];
	int readoffset;
	int writeoffset;
	int bufsize;
	int shrinkbuf;
	char* buffer;
};

#define GG_FLAG_DCCINCOMING     0x0001 
#define GG_FLAG_DCCOUTNATTED    0x0002
#define GG_FLAG_LOGGED          0x0004
#define GG_FLAG_WRITEFORREADING 0x0008
#define GG_FLAG_READFORWRITING  0x0010
#define GG_FLAG_SSL             0x0020
#define GG_FLAG_WAITFORME       0x0040
#define GG_FLAG_CLIENTCONN      0x0080

extern struct list_head gg_fds;
extern pthread_mutex_t gg_fds_mutex;
extern struct gg_fd* gg_fd_mapping[GG_MAXFD];

static __inline__ struct gg_fd* gg_alloc_fd(int fd, int bufsize)
{
	struct gg_fd* f;
	char* buf;

	if (fd >= GG_MAXFD)
	{
		gg_error("error: out of file descriptors, dropping fd %d", fd);
		return NULL;
	}
	
	if (!(f = (struct gg_fd*)malloc(sizeof(*f))))
	{
		gg_error("fatal: cannot allocate fd memory in gg_alloc_fd()");
		return NULL;
	}

	if (!(buf = (char*)malloc(bufsize)))
	{
		gg_error("fatal: cannot allocate buffer memory in gg_alloc_fd()");
		free(f);
		return NULL;
	}

	gg_fd_mapping[fd] = f;
	f->buffer = buf;
	f->refcount = 1;
	f->fd = fd;
	INIT_LIST_HEAD(&f->list);
	f->twin = NULL;
	f->readoffset = 0;
	f->writeoffset = 0;
	f->ip = 0;
	f->uin = 0;
	f->nat_ip = 0;
	f->nat_port = 0;
	f->parent = NULL;
	f->cexp = NULL;
	f->expby = NULL;
	f->injectlen = 0;
	f->flags = 0;
#ifdef HAVE_OPENSSL
	f->ssl = NULL;
#endif
	f->bufsize = bufsize;
	f->shrinkbuf = 0;
	return f;
}

static __inline__ void gg_get_fd(struct gg_fd* f)
{
	f->refcount++;
}

static __inline__ void gg_put_fd(struct gg_fd* f)
{
	if (--(f->refcount) == 0)
	{
		gg_fd_mapping[f->fd] = NULL;
		free(f->buffer);
		free(f);
	}
}

static __inline__ void __gg_attach_fd(struct gg_fd* f)
{
	list_add(&f->list, &gg_fds);
}

static __inline__ void gg_attach_fd(struct gg_fd* f)
{
	pthread_mutex_lock(&gg_fds_mutex);
	__gg_attach_fd(f);
	pthread_mutex_unlock(&gg_fds_mutex);
}

static __inline__ void __gg_detach_fd(struct gg_fd* f)
{
	list_del(&f->list);
}

static __inline__ void gg_detach_fd(struct gg_fd* f)
{
	pthread_mutex_lock(&gg_fds_mutex);
	__gg_detach_fd(f);
	pthread_mutex_unlock(&gg_fds_mutex);
}

struct gg_hashtbl
{
	int size;
	int (*fn)(void*, int);
	int (*cmp)(void*, void*);
	struct list_head heads[0];
};

#define GG_UINH_SIZE 1024
#define GG_UINH_LIMIT 10000

struct gg_uinh_entry
{
	struct list_head list;
	uint32_t uin;
	uint32_t ip;
	uint16_t port;
	struct gg_fd* fd;
};

static __inline__ struct gg_hashtbl* gg_alloc_hashtbl(int size,
		int (*fn)(void*, int), int (*cmp)(void*, void*))
{
	struct gg_hashtbl* h = (struct gg_hashtbl*)
		malloc(sizeof(*h) + size*sizeof(struct gg_hashtbl));
	int i;
	if (!h)
	{
		gg_error("fatal: cannot allocate memory in gg_alloc_hashtbl()");
		return NULL;
	}
	for (i=0; i<size; i++)
		INIT_LIST_HEAD(&h->heads[i]);
	h->size = size;
	h->fn = fn;
	h->cmp = cmp;
	return h;
}

static __inline__ void gg_add_hitem(struct gg_hashtbl* h, void* item)
{
	int bucket = h->fn(item, h->size);
	list_add_tail(item, &h->heads[bucket]);
}

static __inline__ void* gg_find_hitem(struct gg_hashtbl* h, void* item)
{
	int bucket = h->fn(item, h->size);
	struct list_head* i;
	list_for_each(i, &h->heads[bucket])
		if (!h->cmp(i, item))
			return i;
	return NULL;
}

static __inline__ void gg_del_hitem(struct gg_hashtbl* h, void* item)
{
	list_del(item);
	INIT_LIST_HEAD((struct list_head*)item);
}

extern struct list_head gg_fds;
extern pthread_mutex_t gg_fds_mutex;
extern pthread_t gg_main_thread;
extern struct gg_server_desc* gg_servers;
extern struct gg_server_desc* gg_servers_iter;
extern pthread_mutex_t gg_servers_mutex;
extern uint16_t gg_server_port;
extern uint16_t gg_dcc_port;
extern uint32_t gg_outer_ip;
extern char* gg_outer_iface;
extern int gg_max_chat_conns;
extern int gg_max_dcc_conns;
extern int gg_chat_listener;
extern int gg_dcc_listener;
extern int gg_leave_local;
extern int gg_cascade;
extern int gg_dynamic_buffers;

/* main.c */
void gg_kick_main_thread(void);
uint32_t gg_ifname_to_ip(const char* ifname);

/* conn.c */
void gg_close_fd(struct gg_fd* f);
void gg_abort_fd(struct gg_fd* f);

int gg_setup_listener(uint32_t ip, uint16_t port);
int gg_accept_dcc_connection(void);
int gg_accept_chat_connection(int lfd, int ssl);
void* gg_establish_dcc(void* __info);
int gg_need_dcc_server(void);

int gg_fd_wants_reading(struct gg_fd* f);
int gg_fd_wants_writing(struct gg_fd* f);

void gg_fd_handle_reading(struct gg_fd* f);
void gg_fd_handle_writing(struct gg_fd* f);

/* masq.c */
int gg_masq_init(void);

void gg_masq_init_chat_req(struct gg_fd* f);
void gg_masq_init_chat_resp(struct gg_fd* f);
void gg_masq_init_dcc_req(struct gg_fd* f);

void gg_masq_close_handler(struct gg_fd* f);

void gg_masq_timeout_exps(void);

#ifdef HAVE_OPENSSL
/* ssl.c */
int gg_ssl_init(const char* key, const char* cert);
int gg_ssl_accept_connection(struct gg_fd* f);
int gg_ssl_write(struct gg_fd* f);
int gg_ssl_read(struct gg_fd* f);
void gg_ssl_shutdown(struct gg_fd* f, int mode);
void gg_ssl_close(struct gg_fd* f);
#endif

/* GG protocol definitions */

struct gg_header
{
	uint32_t type;
	uint32_t length;
} __attribute__ ((packed));

#define GG_LOGIN 0x000c
#define GG_LOGIN_OK 0x0003

struct gg_login {
	uint32_t uin;
	uint32_t hash;
	uint32_t status;
	uint32_t version;
	uint32_t local_ip;
	uint16_t local_port;
} __attribute__ ((packed));

#define GG_LOGIN_EXT 0x0013

struct gg_login_ext {
	uint32_t uin;
	uint32_t hash;
	uint32_t status;
	uint32_t version;
	uint32_t local_ip;
	uint16_t local_port;
	uint32_t external_ip;
	uint16_t external_port;
} __attribute__ ((packed));

#define GG_LOGIN60 0x0015

struct gg_login60
{
	uint32_t uin;
	uint32_t hash;
	uint32_t status;
	uint32_t version;
	uint8_t d1;
	uint32_t local_ip;
	uint16_t local_port;
	uint32_t external_ip;
	uint16_t external_port;
	uint8_t image_size;
	uint8_t d2;
} __attribute__ ((packed));

#define GG_STATUS_NOT_AVAIL       0x0001
#define GG_STATUS_NOT_AVAIL_DESCR 0x0015
#define GG_STATUS_AVAIL           0x0002
#define GG_STATUS_AVAIL_DESCR     0x0004
#define GG_STATUS_BUSY            0x0003
#define GG_STATUS_BUSY_DESCR      0x0005
#define GG_STATUS_INVISIBLE       0x0014
#define GG_STATUS_INVISIBLE_DESCR 0x0016
#define GG_STATUS_BLOCKED         0x0006

#define GG_STATUS_FRIENDS_MASK 0x8000		

#define GG_S(x) ((x) & ~GG_STATUS_FRIENDS_MASK)

#define GG_S_A(x) (GG_S(x) == GG_STATUS_AVAIL || \
		GG_S(x) == GG_STATUS_AVAIL_DESCR)
	
#define GG_S_NA(x) (GG_S(x) == GG_STATUS_NOT_AVAIL || \
		GG_S(x) == GG_STATUS_NOT_AVAIL_DESCR)
	
#define GG_S_D(x) (GG_S(x) == GG_STATUS_NOT_AVAIL_DESCR || \
		GG_S(x) == GG_STATUS_AVAIL_DESCR || \
		GG_S(x) == GG_STATUS_BUSY_DESCR || \
		GG_S(x) == GG_STATUS_INVISIBLE_DESCR)

#define GG_NOTIFY_REPLY 0x000c	
	
struct gg_notify_reply {
	uint32_t uin;			
	uint32_t status;		
	uint32_t remote_ip;		
	uint16_t remote_port;		
	uint32_t version;		
	uint16_t dunno2;		
} __attribute__ ((packed));

#define GG_NOTIFY_REPLY60 0x0011
	
struct gg_notify_reply60 {
	uint32_t uin;			
	uint8_t status;			
	uint32_t remote_ip;		
	uint16_t remote_port;		
	uint8_t version;		
	uint8_t image_size;		
	uint8_t dunno1;			
} __attribute__ ((packed));

#define GG_STATUS60 0x000f
	
struct gg_status60 {
	uint32_t uin;			
	uint8_t status;			
	uint32_t remote_ip;		
	uint16_t remote_port;		
	uint8_t version;		
	uint8_t image_size;		
	uint8_t dunno1;			
} __attribute__ ((packed));

#define GG_CLASS_CTCP 0x0010

#define GG_SEND_MSG 0x000b

struct gg_send_msg_pkt {
	struct gg_header hdr;
	uint32_t recipient;
	uint32_t seq;
	uint32_t msgclass;
	char     dunno;
} __attribute__ ((packed));

#define GG_RECV_MSG 0x000a

struct gg_recv_msg_pkt {
	struct gg_header hdr;
	uint32_t sender;
	uint32_t seq;
	uint32_t time;
	uint32_t msgclass;
	char     dunno;
} __attribute__ ((packed));

#define GG_DCC_WANT_FILE 0x0003
#define GG_DCC_CATCH_FILE 0x0002

struct gg_dcc_handshake
{
	uint32_t my_uin;
	uint32_t your_uin;
} __attribute__ ((packed));

#define GG_MAX_PKT_TYPE 1000
#define GG_MAX_PKT_LEN  65535

#endif
