/* 
   main.c - simple startup

   Copyright (C) Szymon Acedanski 2003
*/

/*
    ggrelay - Gadu-Gadu instant messager relaying agent
    Copyright (C) 2003 Szymon Acedanski

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <syslog.h>

#if HAVE_POLL
#if HAVE_POLL_H
#include <poll.h>
#ifndef POLLRDNORM
#define POLLRDNORM POLLIN
#endif
#ifndef POLLWRNORM
#define POLLWRNORM POLLOUT
#endif
#else
#undef HAVE_POLL
#endif
#endif

#include "ggrelay.h"

struct gg_server_desc* gg_servers = NULL;
struct gg_server_desc* gg_servers_iter;
pthread_mutex_t gg_servers_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_t gg_main_thread;
LIST_HEAD(gg_fds);
pthread_mutex_t gg_fds_mutex = PTHREAD_MUTEX_INITIALIZER;
uint16_t gg_server_port = GG_DEFAULT_SERVER_PORT;
uint16_t gg_dcc_port = GG_DEFAULT_DCC_PORT;
int gg_terminate = 0;
uint32_t gg_inner_ip = 0;
uint32_t gg_outer_ip = 0;
char* gg_outer_iface = NULL;
int gg_hup_caught = 0;
int gg_logging = 0;
int gg_max_chat_conns = -1;
int gg_max_dcc_conns = -1;
int gg_chat_listener = -1;
int gg_dcc_listener = -1;
int gg_leave_local = 1;
int gg_cascade = 0;
int gg_dynamic_buffers = 1;
struct gg_fd* gg_fd_mapping[GG_MAXFD];
int gg_use_select = 0;

#ifdef HAVE_OPENSSL
uint16_t gg_sslsvr_port = GG_DEFAULT_SSLSVR_PORT;
int gg_ssl_listener = -1;
#endif

static int fake_sockfd = -1;

static void usage(void)
{
	fprintf(stderr,
			"usage: ggrelay [options]\n\n"
			"  -i <ip/if>, --inner-ip <ip/if>  internal IP/interface\n"
			"  -o <ip/if>, --outer-ip <ip/if>  external IP/interface (mandatory)\n"
			"  -d, --daemonize                 go to background after initialization\n"
			"  -s <ip>, --server <ip>          add GG server\n"
			"  -u <user>, --user <user>        suid to specified user at startup\n"
			"  -m <num>, --gg-limit <num>      max. number of chat connections\n"
			"  -n <num>, --dcc-limit <num>     max. number of DCC connections\n"
			"  --static-buffers                disable dynamic buffers allocation\n"
			"  --gg-port                       override chat port (default is %d)\n"
			"  --dcc-port                      override DCC port (default is %d)\n"
			"  --relay-local                   catch also local-network connections\n"
			"  --cascade                       turn on cascading-compatibility\n"
			"  --use-select                    use select() instead of poll()\n"
#ifdef HAVE_OPENSSL
			"  --ssl-port                      override SSL server port (default is %d)\n"
			"  --ssl-key <file>                SSL server private key\n"
			"  --ssl-cert <file>               SSL server certificate\n"
#endif
			"  -l, --logging                   enable connections logging\n"
			"  -v, --verbose                   repeat all messages to stderr, unmask debug messages\n"
			"  -h, --help, --usage             show usage information\n"
			"  --version                       show version information\n\n"
			"Bug reports to <accek@poczta.gazeta.pl>\n",
			GG_DEFAULT_SERVER_PORT, GG_DEFAULT_DCC_PORT
#ifdef HAVE_OPENSSL
			, GG_DEFAULT_SSLSVR_PORT
#endif
			);
	exit(0);
}

static void version(void)
{
	fprintf(stderr,
			"ggrelay version " GGRELAY_VERSION ", Copyright (C) 2003 Szymon Acedanski\n"
			"ggrelay comes with ABSOLUTELY NO WARRANTY; see COPYING for details.\n"
			"This is free software, and you are welcome to reedistribute it\n"
			"under certain conditions; see COPYING for details.\n\n"
			"Bug reports to <accek@poczta.gazeta.pl>\n");
	exit(0);
}

static int add_server(const char* param)
{
	struct gg_server_desc* s = (struct gg_server_desc*)malloc(sizeof(*s));
	char* str;
	char* port;
	if (!s) return 1;
	if (!(str = strdup(param)))
	{
		free(s);
		return 1;
	}
	port = strchr(str, ':');
	DEBUGP("registering server '%s'", param);
	s->next = gg_servers;
	s->addr.sin_family = AF_INET;
	if (port)
	{
		*port = '\0';
		s->addr.sin_port = htons(atoi(port+1));
	}
	else
		s->addr.sin_port = htons(GG_DEFAULT_SERVER_PORT);
	inet_aton(str, &s->addr.sin_addr);
	gg_servers = s;
	free(str);
	return 0;
}

uint32_t gg_ifname_to_ip(const char* ifname)
{
	struct ifreq ifr;
	if (strlen(ifname) >= IFNAMSIZ)
		return 0;
	strcpy(ifr.ifr_name, ifname);
	if (ioctl(fake_sockfd, SIOCGIFADDR, &ifr) < 0)
		return 0;
	return ((struct sockaddr_in*)&ifr.ifr_addr)->sin_addr.s_addr;
}

char* gg_get_default_ifname(void)
{
	static char buff[1024];
	struct ifconf ifc;

	ifc.ifc_len = sizeof(buff);
	ifc.ifc_buf = buff;

	if (ioctl(fake_sockfd, SIOCGIFCONF, &ifc) < 0)
	{
		gg_warn("Could not get interface list");
		return NULL;
	}

	if (ifc.ifc_len < sizeof(struct ifreq))
	{
		gg_warn("warning: No interfaces configured");
		return NULL;
	}
	
	return ifc.ifc_req->ifr_name;
}

#ifdef HAVE_POLL
static struct pollfd pfds[GG_MAXFD];
static int nfds;

static inline void add_to_pfds(int fd, int flags)
{
	pfds[nfds].fd = fd;
	pfds[nfds].events = flags;
	pfds[nfds].revents = 0;
	nfds++;
}

static void poll_loop(void)
{
	DEBUGP("entering poll() loop ...");
	while (!gg_terminate)
	{
		struct gg_fd* f, * j;
		int i;
		int err;
		
		do
		{
			nfds = 0;

			pthread_mutex_lock(&gg_fds_mutex);
			gg_hup_caught = 0;
			if (gg_chat_listener >= 0)
				add_to_pfds(gg_chat_listener, POLLRDNORM);
			if (gg_dcc_listener >= 0)
				add_to_pfds(gg_dcc_listener, POLLRDNORM);
#ifdef HAVE_OPENSSL
			if (gg_ssl_listener >= 0)
				add_to_pfds(gg_ssl_listener, POLLRDNORM);
#endif
			list_for_each_entry_safe(f, j, &gg_fds, list)
			{
				if (f->state == GG_S_DESTROY)
				{
					__gg_detach_fd(f);
					gg_put_fd(f);
					continue;
				}
				add_to_pfds(f->fd, POLLERR |
						(gg_fd_wants_reading(f) ? POLLRDNORM : 0) |
						(gg_fd_wants_writing(f) ? POLLWRNORM : 0));
			}
			pthread_mutex_unlock(&gg_fds_mutex);
		} while (gg_hup_caught);

		err = poll(pfds, nfds, -1);
		if (err == -1)
		{
			if (errno == EINTR)
				continue;
			gg_error("fatal: poll() failed: %s", strerror(errno));
			exit(1);
		}

		pthread_mutex_lock(&gg_fds_mutex);
		i = 0;
		if (gg_chat_listener >= 0 && pfds[i++].revents)
			gg_accept_chat_connection(gg_chat_listener, 0);
#ifdef HAVE_OPENSSL
		if (gg_ssl_listener >= 0 && pfds[i++].revents)
			gg_accept_chat_connection(gg_ssl_listener, 1);
#endif
		if (gg_dcc_listener >= 0 && pfds[i++].revents)
			gg_accept_dcc_connection();

		for (; i<nfds; i++)
		{
			struct gg_fd* f;
			if (!pfds[i].revents)
				continue;
			f = gg_fd_mapping[pfds[i].fd];
			gg_assert(f != NULL);
			if (pfds[i].revents & POLLERR)
			{
				gg_abort_fd(f);
				continue;
			}
			if (pfds[i].revents & POLLRDNORM)
			{
#ifdef HAVE_OPENSSL
				if (f->flags & GG_FLAG_READFORWRITING)
					gg_fd_handle_writing(f);
				else
#endif
					gg_fd_handle_reading(f);
			}
			if (pfds[i].revents & POLLWRNORM)
			{
#ifdef HAVE_OPENSSL
				if (i->flags & GG_FLAG_WRITEFORREADING)
					gg_fd_handle_reading(f);
				else
#endif
					gg_fd_handle_writing(f);
			}
		}
		
		gg_masq_timeout_exps();
		pthread_mutex_unlock(&gg_fds_mutex);
	}
}
#endif /* HAVE_POLL */

static inline void add_to_fdset(fd_set* s, int fd, int* max)
{
	FD_SET(fd, s);
	if (fd > *max)
		*max = fd;
}

static void select_loop(void)
{
	DEBUGP("entering select() loop ...");
	while (!gg_terminate)
	{
		fd_set rd, wr;
		struct gg_fd* i, * j;
		int maxfd = 0;
		int err;
		
		do
		{
			FD_ZERO(&rd);
			FD_ZERO(&wr);

			pthread_mutex_lock(&gg_fds_mutex);
			gg_hup_caught = 0;
			if (gg_chat_listener >= 0)
				add_to_fdset(&rd, gg_chat_listener, &maxfd); 
#ifdef HAVE_OPENSSL
			if (gg_ssl_listener >= 0)
				add_to_fdset(&rd, gg_ssl_listener, &maxfd); 
#endif
			if (gg_dcc_listener >= 0)
				add_to_fdset(&rd, gg_dcc_listener, &maxfd); 

			list_for_each_entry_safe(i, j, &gg_fds, list)
			{
				if (i->state == GG_S_DESTROY)
				{
					__gg_detach_fd(i);
					gg_put_fd(i);
					continue;
				}
				if (gg_fd_wants_reading(i)) add_to_fdset(&rd, i->fd, &maxfd);
				if (gg_fd_wants_writing(i)) add_to_fdset(&wr, i->fd, &maxfd);
			}
			pthread_mutex_unlock(&gg_fds_mutex);
		} while (gg_hup_caught);

		err = select(maxfd+1, &rd, &wr, NULL, NULL);
		if (err == -1)
		{
			if (errno == EINTR)
				continue;
			gg_error("fatal: select() failed: %s", strerror(errno));
			exit(1);
		}

		pthread_mutex_lock(&gg_fds_mutex);
		list_for_each_entry(i, &gg_fds, list)
		{
			if (FD_ISSET(i->fd, &rd))
			{
#ifdef HAVE_OPENSSL
				if (i->flags & GG_FLAG_READFORWRITING)
					gg_fd_handle_writing(i);
				else
#endif
					gg_fd_handle_reading(i);
			}
			if (FD_ISSET(i->fd, &wr))
			{
#ifdef HAVE_OPENSSL
				if (i->flags & GG_FLAG_WRITEFORREADING)
					gg_fd_handle_reading(i);
				else
#endif
					gg_fd_handle_writing(i);
			}
		}
		if (gg_chat_listener >= 0 && FD_ISSET(gg_chat_listener, &rd))
			gg_accept_chat_connection(gg_chat_listener, 0);
#ifdef HAVE_OPENSSL
		if (gg_ssl_listener >= 0 && FD_ISSET(gg_ssl_listener, &rd))
			gg_accept_chat_connection(gg_ssl_listener, 1);
#endif
		if (gg_dcc_listener >= 0 && FD_ISSET(gg_dcc_listener, &rd))
			gg_accept_dcc_connection();

		gg_masq_timeout_exps();
		pthread_mutex_unlock(&gg_fds_mutex);
	}
}

static void huphandler(int sig)
{
	gg_hup_caught = 1;
}

static void termhandler(int sig)
{
	gg_terminate = 1;
}

void gg_kick_main_thread(void)
{
	pthread_kill(gg_main_thread, SIGHUP);
}

int main(int argc, char** argv)
{
	int run_background = 0;
	int gg_logging_flags = 0;
	int gg_logging_mask = 0xffffffff & ~LOG_MASK(LOG_DEBUG);
	int setuid_uid = -1;
#ifdef HAVE_OPENSSL
	const char* ssl_key = NULL;
	const char* ssl_cert = NULL;
#endif
	
	if ((fake_sockfd = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
	{
		fprintf(stderr, "fatal: cannot create fake socket\n");
		return 1;
	}
	
	while (1)
	{
		int option_index = 0;
		char c;
		static struct option long_options[] =
		{
			{"inner-ip", 1, 0, 'i'},
			{"outer-ip", 1, 0, 'o'},
			{"daemonize", 0, 0, 'd'},
			{"server", 1, 0, 's'},
			{"gg-port", 1, 0, 2},
			{"dcc-port", 1, 0, 1},
			{"gg-limit", 1, 0, 'm'},
			{"dcc-limit", 1, 0, 'n'},
			{"static-buffers", 0, 0, 9},
			{"user", 1, 0, 'u'},
			{"logging", 0, 0, 'l'},
			{"relay-local", 0, 0, 7},
			{"cascade", 0, 0, 8},
			{"use-select", 0, 0, 10},
#ifdef HAVE_OPENSSL
			{"ssl-port", 1, 0, 6},
			{"ssl-key", 1, 0, 4},
			{"ssl-cert", 1, 0, 5},
#endif
			{"verbose", 0, 0, 'v'},
			{"help", 0, 0, 'h'},
			{"usage", 0, 0, 'h'},
			{"version", 0, 0, 3},
			{0, 0, 0, 0}
		};

		c = getopt_long (argc, argv, "i:o:hds:u:vlm:n:y", long_options, &option_index);
		if (c == -1) break;

		switch (c)
		{
			case 'i':
				if (!inet_aton(optarg, (struct in_addr*)&gg_inner_ip) &&
						!(gg_inner_ip = gg_ifname_to_ip(optarg)))
				{
					fprintf(stderr, "fatal: inner relay address is invalid\n");
					return 1;
				}
				break;

			case 'o':
				if (!inet_aton(optarg, (struct in_addr*)&gg_outer_ip))
				{
					gg_outer_iface = optarg;
					gg_outer_ip = 0;
					if (!gg_ifname_to_ip(optarg))
						fprintf(stderr, "warning: -o parameter may be invalid\n");
				}
				break;
				
			case 'd':
				run_background = 1;
				break;

			case 's':
				add_server(optarg);
				break;

			case 1:
				gg_dcc_port = atoi(optarg);
				break;

			case 2:
				gg_server_port = atoi(optarg);
				break;

			case 'm':
				gg_max_chat_conns = atoi(optarg);
				break;

			case 'n':
				gg_max_dcc_conns = atoi(optarg);
				break;

			case 9:
				gg_dynamic_buffers = 0;
				break;

			case 'u':
			{
				struct passwd* p = getpwnam(optarg);
				if (!p)
				{
					fprintf(stderr, "fatal: option --user: no such user\n");
					return 1;
				}
				setuid_uid = p->pw_uid;
				break;
			}

			case 'l':
				gg_logging = 1;
				break;

			case 'v':
				gg_logging_flags |= LOG_PERROR;
				gg_logging_mask = 0xffffffff;
				break;

			case 7:
				gg_leave_local = 0;
				break;

			case 8:
				gg_cascade = 1;
				break;
			
			case 10:
				gg_use_select = 1;
				break;

#ifdef HAVE_OPENSSL
			case 6:
				gg_sslsvr_port = atoi(optarg);
				break;

			case 4:
				ssl_key = strdup(optarg);
				break;

			case 5:
				ssl_cert = strdup(optarg);
				break;
#endif

			case 'h':
				usage();
				break;

			case 3:
				version();
				break;

			case '?':
				fprintf(stderr, "error: unknown command-line option\n");
				exit(1);
				break;

			case ':':
				fprintf(stderr, "error: missing argument\n");
				exit(1);
				break;
		}
	}

	if (optind < argc)
		fprintf(stderr, "warning: additional command-line parameters ignored\n");

	openlog("ggrelay", gg_logging_flags, LOG_DAEMON);
	gg_notice("ggrelay v. " GGRELAY_VERSION " starting up");
	signal(SIGHUP, huphandler);
	signal(SIGTERM, termhandler);
	signal(SIGINT, termhandler);
	signal(SIGPIPE, SIG_IGN);

	gg_main_thread = pthread_self();
	if (gg_servers == NULL)
		add_server(GG_DEFAULT_SERVER);
	gg_servers_iter = gg_servers;
	
	if (!gg_outer_ip && !gg_outer_iface)
	{
		if (!(gg_outer_iface = gg_get_default_ifname()))
		{
			gg_error("fatal: external interface not specified, please use -o option");
			return 1;
		}
		else
			gg_warn("warning: assuming default interface '%s' is external one",
					gg_outer_iface);
	}
	
#ifdef HAVE_OPENSSL
	if (ssl_key && ssl_cert)
	{
		if (gg_ssl_init(ssl_key, ssl_cert))
			return 1;
		if ((gg_ssl_listener = gg_setup_listener(gg_inner_ip, gg_sslsvr_port)) < 0)
		{
			gg_error("fatal: cannot establish ssl listener");
			return 1;
		}
	}
#endif
	
	if ((gg_chat_listener = gg_setup_listener(gg_inner_ip, gg_server_port)) < 0)
	{
		gg_error("fatal: cannot establish chat listener");
		return 1;
	}
	
	if (!gg_outer_iface && (gg_dcc_listener =
				gg_setup_listener(gg_outer_ip, gg_dcc_port)) < 0)
	{
		gg_error("fatal: cannot establish DCC listener");
		return 1;
	}
	else if (gg_outer_iface)	/* don't be lazy, someone can be faster */
		gg_need_dcc_server();
	
	if (setuid_uid != -1 && setuid(setuid_uid))
	{
		gg_error("fatal: cannot suid");
		return 1;
	}

	if (gg_masq_init())
		return 1;

	if (run_background)
	{
		gg_notice("initialized, going into background");
		daemon(0, 0);
	}

#ifdef HAVE_POLL
	if (!gg_use_select)
		poll_loop();
	else
#endif
		select_loop();

	gg_notice("terminated.");
	closelog();
	return 0;
}
